NETWORKS = {
    "abstract": [
        "testnet",
    ],
    "apechain": [
        "mainnet",
        "curtis",
    ],
    "arbitrum": [
        "mainnet",
        "nova",
        "sepolia",
    ],
    "astar": [
        "mainnet",
    ],
    "avalanche": [
        "mainnet",
        "fuji",
    ],
    "base": [
        "mainnet",
        "sepolia",
    ],
    "berachain": [
        "bartio",
    ],
    "blast": [
        "mainnet",
        "sepolia",
    ],
    "bsc": [
        "mainnet",
        "testnet",
        "opbnb",
        "opbnb-testnet",
    ],
    "crossfi": [
        "mainnet",
        "testnet",
    ],
    "ethereum": [
        "holesky",
        "mainnet",
        "sepolia",
    ],
    "fantom": [
        "opera",
        "testnet",
    ],
    "flow-evm": [
        "mainnet",
        "testnet",
    ],
    "geist": [
        "mainnet",
        "polter",
    ],
    "gnosis": [
        "mainnet",
        "chiado",
    ],
    "lens": [
        "sepolia",
    ],
    "linea": [
        "mainnet",
        "sepolia",
    ],
    "lumia": [
        "prism",
        "testnet",
    ],
    "mantle": [
        "mainnet",
        "sepolia",
    ],
    "metis": [
        "mainnet",
    ],
    "optimism": [
        "mainnet",
        "sepolia",
    ],
    "polygon": [
        "mainnet",
        "amoy",
    ],
    "polygon-zkevm": [
        "mainnet",
        "cardona",
    ],
    "rootstock": [
        "mainnet",
        "testnet",
    ],
    "scroll": [
        "mainnet",
        "sepolia",
    ],
    "shape": [
        "mainnet",
        "sepolia",
    ],
    "soneium": [
        "minato",
    ],
    "unichain": [
        "sepolia",
    ],
    "xmtp": [
        "sepolia",
    ],
    "world-chain": [
        "mainnet",
        "sepolia",
    ],
    "zetachain": [
        "mainnet",
        "testnet",
    ],
    "zksync": [
        "mainnet",
        "sepolia",
    ],
    "zora": [
        "mainnet",
        "sepolia",
    ],
}
